#!/usr/bin/perl

# SWeBLAST.pl - Cut sequences in a series of windows and BLAST them.
#
# Written: June 4, 2007
# Author:  Mathieu Fourment, Macquarie University
#
#Description
#   -help      or -h :  Show this message.
#   -input     or -i :  Input file.
#   -output    or -o :  Output directory to be created with the results files (OPTIONAL).
#   -window    or -w :  Window length default 100 (OPTIONAL)
#   -step      or -s :  Step length - default 50 (OPTIONAL)
#   -database  or -d :  Database - default nr (OPTIONAL)
#   -program   or -p :  Program: blastn, blastp, blastx - default blastn (OPTIONAL)
#   -expect    or -e :  EXPECT (OPTIONAL)
#   -entrez    or -q :  ENTREZ QUERY (OPTIONAL)
#
#   -parse     or -r :  Call SWeBLAST_parser.pl (OPTIONAL).
#   -cutoff    or -c :  Hits with an e-value above the specified value will be discarded (OPTIONAL).
#
#Output
#   blast.txt       :  BLAST file result
#   blast.txt_STAT  :  Directory with files containg stats
#   seq.fas         :  FASTA file containing the sequences cut according to the window and step
#
#Exemple
#   perl SWeBLAST.pl -p blastn -d nr -i dna.fasta -o myRESULT -w 100 -s 50\n -parse -c 0.001\n";
#     or
#   perl SWeBLAST.pl
 

use strict;
use warnings;
use HTTP::Request::Common;
use Getopt::Long;
use LWP::UserAgent;
use File::Spec;

my $ua = LWP::UserAgent->new;

my $numArgs = $#ARGV + 1;

my $help;
my $input;
my $output;
my $program;
my $database;
my $window = 100;
my $step   = 50;
my $cutoff;
my $parse;
my $entrez_query;
my $expect;

my $URLBASE = 'http://www.ncbi.nlm.nih.gov/blast/Blast.cgi';

my %PUT = (
	'CMD'      => 'Put',
	'DATABASE' => 'nr',
	'PROGRAM'  => 'blastn',

	#'ENTREZ_QUERY' => 'virus NOT influenza [ALL]',
);

my %GET = (
	'CMD'           => 'Get',
	'FORMAT_OBJECT' => 'SearchInfo',
);

my %RETRIEVAL = (
	'CMD'         => 'Get',
	'FORMAT_TYPE' => 'Text',
);

my %prog = (
	blastp    => "blastp",
	blastn    => "blastn",
	blastx    => "blastx",
	megablast => "blastn&MEGABLAST=on",
	rpsblast  => "blastp&SERVICE=rpsblast"
);

if ( $numArgs == 0 ) {
	interactive();
}
else {
	GetOptions(
		'help|h'       => \$help,
		'input|i=s'    => \$input,
		'output|o=s'   => \$output,
		'program|p=s'  => \$program,
		'database|b=s' => \$database,
		'window|w=i'   => \$window,
		'step|s=i'     => \$step,
		'cutoff|c=f'   => \$cutoff,
		'parse|r'      => \$parse,
		'expect|e=f'   => \$expect,
		'entrez|q=s'   => \$entrez_query,
	);
}

if ( defined $program and not defined $prog{$program} ) {
	print STDERR "Program name not correct: $program\n";
	exit 0;
}

my @dirs = File::Spec->splitdir($0);
pop @dirs;
my $parser = File::Spec->catfile( @dirs, "SWeBLAST_parser.pl" );

unless ( -e $parser ) {
	print STDERR "SWeBLAST_parser.pl not found in the path\n";
	exit 3;
}

$PUT{'DATABASE'}     = $database     if ( defined $database );
$PUT{'PROGRAM'}      = $program      if ( defined $program );
$PUT{'ENTREZ_QUERY'} = $entrez_query if ( defined $entrez_query );
$PUT{'EXPECT'}       = $expect       if ( defined $expect );

###############################
#####    TEST
#$input = "/home/mathieu/Desktop/SWeBLAST/test.fa";
#$parse=1;
#$window   = 200;
#$step     = 200;
###############################
help() if ($help);

help("No input file specified") if ( not defined $input );

# check output
$output = "$input" . "_BLAST" if ( not defined $output );
$output =~ s/[\\\/]?$//;

mkdir( $output, 0755 ) || die "Cannot mkdir $output $!" if ( not -e $output );

my $now = time();

#$program = $prog{$program};

# read and encode the queries
my %sequence;
my $headFasta = "";
my $seq       = "";
open( QUERY, $input ) or die "Cannot open $input, $!\n";
while (<QUERY>) {
	$_ =~ s/\r?\n$//;
	if ( $_ =~ /^>/ & $seq ne "" ) {
		$sequence{$headFasta} = $seq;
		$headFasta            = $_;
		$seq                  = "";
	}
	elsif ( $_ =~ /^>/ ) {
		$headFasta = $_;
	}
	else {
		$seq = $seq . $_;
	}
}
$sequence{$headFasta} = $seq;
close(QUERY);

my $output_blast     = File::Spec->catfile( $output, "blast.txt" );
my $output_sequences = File::Spec->catfile( $output, "seq.fa" );

open( OUTPUT, ">$output_blast" )     or die "Cannot open $output, $!\n";
open( SEQ,    ">$output_sequences" ) or die "Cannot open seq.fas, $!\n";

for my $header ( keys %sequence ) {
	if ( $step eq "" or $window eq "" ) {
		my $query = $header . "\n" . $sequence{$header};
		lookup($query);
	}
	else {
		my $i = 0;
		while ( $i + $window < length( $sequence{$header} ) ) {
			my $begin    = $i + 1;
			my $end      = $i + $window;
			my $sequence = substr( $sequence{$header}, $i, $window );
			my $query = $header . '|' . $begin . '-' . $end . "\n" . $sequence;
			print $query. "\n";
			print SEQ $query . "\n";
			lookup($query);

	  #=for comment
	  #			if($program eq "blastx"){
	  #			    $dna2=substr($dna,1);
	  #			    print SEQ $s[0]."|ORF 2|".$temp1."-".$temp2."\n".$dna2."\n";
	  #			    $seq1=uri_escape($s[0]."|ORF 2|".$temp1."-".$temp2."\n".$dna2);
	  #			    lookup($seq1);
	  #			    $dna2=substr($dna2,1);
	  #			    print SEQ $s[0]."|ORF 3|".$temp1."-".$temp2."\n".$dna2."\n";
	  #			    $seq1=uri_escape($s[0]."|ORF 3|".$temp1."-".$temp2."\n".$dna2);
	  #			    lookup($seq1);
	  #			    $revcom=reverse($dna);
	  #			    $revcom=~tr/agtcAGTC/tcagTCAG/;
	  #			    print SEQ $s[0]."|REV 1|".$temp1."-".$temp2."\n".$revcom."\n";
	  #			    $seq1=uri_escape($s[0]."|REV 1|".$temp1."-".$temp2."\n".$revcom);
	  #			    lookup($seq1);
	  #			    $revcom2=substr($revcom,1);
	  #			    print SEQ $s[0]."|REV 2|".$temp1."-".$temp2."\n".$revcom2."\n";
	  #			    $seq1=uri_escape($s[0]."|REV 2|".$temp1."-".$temp2."\n".$revcom2);
	  #			    lookup($seq1);
	  #			    $revcom2=substr($revcom2,1);
	  #			    print SEQ $s[0]."|REV 3|".$temp1."-".$temp2."\n".$revcom2."\n";
	  #			    $seq1=uri_escape($s[0]."|REV 3|".$temp1."-".$temp2."\n".$revcom2);
	  #			    lookup($seq1);
	  #			}
	  #=cut
			$i += $step;
		}    # end while window
	}
}    # end for each sequence
close(OUTPUT);
close(SEQ);

my $after = time();
my $tot   = $after - $now;

if ( $parse or defined $cutoff ) {
	my $cmd = "perl $parser -i $output/blast.txt";
	if ( defined $cutoff ) {
		$cmd .= " -t $cutoff";
	}
	print STDERR "Parsing...";
	`$cmd`;
}
print STDERR " Done\n";
print STDERR "time: $tot\n";

sub lookup {
	my ($seq) = @_;

	# build the request
	my %header = %PUT;
	$header{'QUERY'} = $seq;
	my $req = POST $URLBASE, [%header];

	# get the response
	my $response = $ua->request($req);

	if ( $response->is_success ) {
		if ( $response->content =~
/((:?Nucleotide|Protein) FASTA provided for (:?protein|nucleotide) sequence)/i
		  )
		{
			print STDERR "ERROR $1\n";
			exit 0;
		}

		# parse out the request id
		$response->content =~ /^\s+RID = (.*$)/m;
		my $rid = $1;

		# parse out the estimated time to completion
		$response->content =~ /^\s+RTOE = (.*$)/m;
		my $rtoe = $1;

		# wait for search to complete
		print STDERR "Waiting " . $rtoe . " seconds RID=$rid\n";
		sleep $rtoe;

		# poll for results
		while (1) {
			print STDERR "Waiting " . $rtoe . " seconds\n";

			#sleep $rtoe;
			sleep 5;

#$req = new HTTP::Request GET => "http://www.ncbi.nlm.nih.gov/blast/Blast.cgi?CMD=Get&FORMAT_OBJECT=SearchInfo&RID=$rid";
			%header        = %GET;
			$header{'RID'} = $rid;
			$req           = POST $URLBASE, [%header];
			$response      = $ua->request($req);

			if ( $response->is_success ) {

				if ( $response->content =~ /\s+Status=WAITING/m ) {
					print STDERR "Waiting...\n";
					if ( $response->content =~
/<p class="WAITING">This page will be automatically updated in <b>(\d+)<\/b> seconds<\/p>/
					  )
					{
						$rtoe = $1;
						print STDERR "RTOE $rtoe\n";
					}
					else {
						$rtoe = 10;
					}
					next;
				}

				if ( $response->content =~ /Status=FAILED/m ) {
					print STDERR
"Search $rid failed; please report to blast-help\@ncbi.nlm.nih.gov.\n";
					last;

					#exit 4;
				}

				if ( $response->content =~ /Status=UNKNOWN/m ) {
					print STDERR "Search $rid expired.\n";
					last;

					#exit 3;
				}

				if ( $response->content =~ /Status=READY/m ) {
					if ( $response->content =~ /\s+ThereAreHits=yes/m ) {
						print STDERR "Search complete, retrieving results...\n";
						last;
					}
					else {
						print STDERR "No hits found.\n";
						last;

						#exit 2;
					}
				}

				# if we get here, something unexpected happened.
				print STDERR "Unexpected error\n";
				print $req->as_string() . "\n";
				print $response->content . "\n";
				last;

				#exit 5;

			}    # end if is_success
			else {
				warn(   "request was "
					  . $req->as_string() . "\n"
					  . $response->error_as_HTML );
			}

		}    # end poll loop

		# retrieve and display results
		%header        = %RETRIEVAL;
		$header{'RID'} = $rid;
		$req           = POST $URLBASE, [%header];
		$response      = $ua->request($req);
		if ( $response->is_success ) {
			print OUTPUT ( $response->content . "\n" );
		}
		else {
			warn(   "request was "
				  . $req->as_string() . "\n"
				  . $response->error_as_HTML );
		}
	}
	else {
		warn(   "request was "
			  . $req->as_string() . "\n"
			  . $response->error_as_HTML );
	}
}

sub interactive {
	$input = prompt("Enter the name of the file containing the sequences");
	die "$input cannot be read" unless(-e $input);
	
	$output = prompt("Enter the name of the directory where the results will be stored","$input" . "_BLAST");
	print "$output\n";

	$window = prompt( "Enter the window size", $window );
	print "$window\n";
	
	$step = prompt( "Enter the step size", $step );
	print "$step\n";
	
	$database = prompt( "Enter database to be used", $PUT{'DATABASE'} );
	print "$database\n";
	
	$program = prompt( "Enter the program to be used", $PUT{'PROGRAM'} );
	print "$program\n";
	
	$entrez_query = prompt("Enter a query to restrict the search (OPTIONAL)");
	print "$entrez_query\n" if(defined $entrez_query);
	
	my $doweparse = prompt( "Do you want to parse the results now? (Y/n)", "Y" );
	if ( $doweparse =~ /Y/i ) {
		$parse = 1;
		$cutoff = prompt("Define a cutoff for the e-value (OPTIONAL)");    
	}
}

sub prompt($$) {
	my ( $text, $default ) = @_;
   if ($default) {
      print $text, "[", $default, "]:\n";
   } else {
      print $text, ":\n";
   }
	

	$| = 1;          # force a flush after our print
	$_ = <STDIN>;    # get the input from STDIN (presumably the keyboard)

	s/\r?\n//g;

	if ( defined $default ) {
		return $_ ? $_ : $default;
	}
	else {
		return $_;
	}
}

sub help {
	my ($message) = @_;
	print "\n$message\n\n" if ( defined $message );
	print "Name
    $0 - Cut sequences in a series of windows and BLAST them.
Synopsis
    $0 [OPTION]...
Description
   -help     or -h :  Show this message.
   -input    or -i :  Input file.
   -output   or -o :  Output directory to be created with the results files (OPTIONAL).
   -window   or -w :  Window length default 100 (OPTIONAL)
   -step     or -s :  Step length - default 50 (OPTIONAL)   
   -database or -d :  Database - default nr (OPTIONAL)    
   -program  or -p :  Program: blastn, blastp, blastx - default blastn (OPTIONAL)
   -expect   or -e :  EXPECT default 10 (OPTIONAL)
   -entrez   or -q :  ENTREZ QUERY (OPTIONAL)   
   
   -parse    or -r :  Call SWeBLAST_parser.pl (OPTIONAL).
   -cutoff   or -c :  Hits with an e-value above the specified value will be discarded (OPTIONAL).

Output
   blast.txt       :  BLAST file result
   blast.txt_STAT  :  Directory with files containg stats 
   seq.fas         :  FASTA file containing the sequences cut according to the window and step 

Exemple
   perl $0 -p blastn -d nr -i dan.fasta -o myRESULT -w 100 -s 50\n -parse -c 0.001\n";
	exit 0;
}
