#!/usr/bin/perl

# SWeBLAST_parser.pl - Parse a blast file generated from SWeBLAST.pl and give some statistics (ranking, e-value...) about the file.
#
# Written: June 4, 2007
# Author:  Mathieu Fourment, Macquarie University
#
#Description
#   -help or -h :      Show this message.
#   -input or -i :     Input file.
#   -output or -o :    Output directory to be created with the results files (OPTIONAL).
#   -threshold or -t:  Hits with an e-value above the specified value will be discarded (OPTIONAL).
#
#Output
#   summary.csv :      Display the results
#   stat.csv :         Display some stats on the log of the e-value and rank (min,max,median,mean,standard deviation)
#   evalue.csv :       Display the log of the e-value
#   rank.csv :         Display the rank of the sequences
#   frame.csv :        Display the the orientation of the ORF (+1,+2,+3,-1,-2,-3)
#
#Exemple
#   perl SWeBLAST_parser.pl -i blast1.txt -o /home/user/project1 [-t 0.0001]\n";

use Getopt::Long;
use strict;
use warnings;
use File::Spec;

my $help;
my $input;
my $output;
my $threshold;

my $numArgs = $#ARGV + 1;

if($numArgs==0){
	interactive();
}
else{
	GetOptions(
		'help|h'        => \$help,
		'input|i=s'     => \$input,
		'output|o=s'    => \$output,
		'threshold|t=f' => \$threshold,
	);
	help() if($help);
}


###############################
#####    TEST
#$input = "/home/mathieu/Desktop/SWeBLAST/test.fa_BLAST/blast.txt";
###############################
# check input
help("No input file specified") if ( not defined $input );

# check output
$output = "$input" . "_STAT" if ( not defined($output) );
$output =~ s/[\\\/]$//;

eval { mkdir($output) || die "Cannot create $output: $!"; };
#warn $@ if $@;

my $nb      = -1;
my $ranking = 0;
my ( $program, $seqQuery, $previous, $name, $e, );
my %hits;

my $output_summary=File::Spec->catfile($output, "summary.csv");

open( INPUT, $input ) or die "Cannot open $input, $!\n";
open( OUTPUT, ">$output_summary" )
  or die "Cannot open $output/summary.csv, $!\n";
while (<INPUT>) {
	$_ =~ s/\r?\n$//;

	# Get program: blastn, blastp, blastx
	if (/^(t?blast[pnx])\s.+/i) {
		$program = $1;
	}

	# Get query name
	if ( $_ =~ /^Query=\s+(\S.+)/ ) {
		$seqQuery = $1;
		$nb++;
		$ranking  = 0;
		$previous = "";
		while (<INPUT>) {
			last if (/^Length/);
			$_ =~ s/\r?\n$//;
			$seqQuery .= " " . $_;
		}
		$seqQuery =~ s/,/ /g;
		print OUTPUT $seqQuery;

	}

	# HIT
	elsif ( $_ =~ /^>(.+)/ ) {
		$name = $1;
		while (<INPUT>) {
			$_ =~ s/\r?\n$//;
			last if (/^Length/);
			$name .= " " . $_;
		}
		$name =~ s/,/ /g;

		while (<INPUT>) {
			$_ =~ s/\r?\n$//;
			if (/^\sScore.+=\s(\S+)$/) {
				$e = $1;
				last;
			}
		}

		print OUTPUT "," . $name . "(" . $e . ")";

		$ranking++ if ( $previous eq "" or $previous ne $e ); # In case of a tie

		if ( not defined $threshold or (defined $threshold and $e < $threshold) ) {
			if ( $hits{$name} ) {
				$hits{$name}->{EVALUE}[$nb] = $e;
				$hits{$name}->{RANK}[$nb]   = $ranking;
			}
			else {
				my $rec   = {};
				my @tempe = ();
				$tempe[$nb] = $e;
				my @tempr = ();
				$tempr[$nb]    = $ranking;
				$rec->{EVALUE} = [@tempe];
				$rec->{RANK}   = [@tempr];
				$hits{$name}   = $rec;
			}
		}    # end test threshold
		$previous = $e;
	}

	# Used when no hits found
	elsif ( $_ =~ /<b>No significant similarity found.<\/b>/ ) {
		print OUTPUT ",No hits found";
	}

	# For BLASTX
	elsif ( $_ =~ / Frame = (.+)/ ) {

		# if previously added then it is above the threshold
		if ( $hits{$name} ) {
			$hits{$name}->{FRAME}[$nb] = $1;
		}
	}

	# End of one blast
	elsif (/^  Database/) {
		print OUTPUT "\n";
	}

	# Not sure if it is used
	elsif ( $_ =~ /\*{5,5} No hits found \*{6,6}/ ) {

		#print OUTPUT ",No hits found\n";
	}
}
close(INPUT);

order_log_evalue();
order_rank();

printEvalue();
printRank();
printStat();
printFrame() if ( $program =~ /blastx/i );

close(OUTPUT);

#######################################################################
#                               SUBROUTINES
########################################################################

sub order_log_evalue {
	for my $f ( keys %hits ) {
		my @evalue = @{ $hits{$f}->{EVALUE} };
		for ( my $i = 0 ; $i < scalar @evalue ; $i++ ) {
			if ( not defined $evalue[$i] ) {
				splice( @evalue, $i--, 1 );
			}
			elsif ( $evalue[$i] + 0 == 0 ) {
				$evalue[$i] = 9999;
			}
			else {
				$evalue[$i] = abs( ( log( $evalue[$i] ) / log(10) ) );
			}
		}
		my @evalue_ordered = sort { $a <=> $b } @evalue;
		$hits{$f}->{EVALUEORDERED} = [@evalue_ordered];
	}
}

sub order_rank {
	for my $f ( keys %hits ) {
		my @rank = @{ $hits{$f}->{RANK} };

		for ( my $i = 0 ; $i < scalar @rank ; $i++ ) {
			if ( not defined $rank[$i] ) {
				splice( @rank, $i--, 1 );
			}
		}
		my @rank_ordered = sort { $a <=> $b } @rank;
		$hits{$f}->{RANKORDERED} = [@rank_ordered];
	}
}

sub printStat {
	my (
		@ev,      @r,      $maxe,   $maxr,  $minr,  $mine, $mediane,
		$medianr, $rangee, $ranger, $meane, $meanr, $stde, $stdr
	);
	my $output_stat=File::Spec->catfile($output, "stat.csv");
	open( STAT, ">$output_stat" ) or die "Cannot open $output_stat\n$!\n";
	print STAT "Sequence,nb,min evalue,max evalue,range evalue,median evalue,mean evalue,std evalue,min rank,max rank,range rank,median rank,mean rank,std rank\n";
	for my $f ( keys %hits ) {
		# e-value
		my @evalue = @{ $hits{$f}->{EVALUEORDERED} };

		$mediane = median(@evalue);
		$maxe    = $evalue[0];
		$mine    = $evalue[ scalar @evalue - 1 ];
		$rangee  = abs( $maxe - $mine );
		$meane   = mean(@evalue);
		$stde    = std( $meane, @evalue );

		# Rank
		my @rank = @{ $hits{$f}->{RANKORDERED} };

		$medianr = median(@rank);
		$maxr    = $rank[ scalar @rank - 1 ];
		$minr    = $rank[0];
		$ranger  = abs( $maxr - $minr );
		$meanr   = mean(@rank);
		$stdr    = std( $meanr, @rank );

		print STAT $f . ","
		  . scalar(@evalue) . ","
		  . $mine . ","
		  . $maxe . ","
		  . $rangee . ","
		  . $mediane . ","
		  . $meane . ","
		  . $stde . ","
		  . $minr . ","
		  . $maxr . ","
		  . $ranger . ","
		  . $medianr . ","
		  . $meanr . ","
		  . $stdr . "," . "\n";
	}
	close(STAT);

}

sub printEvalue {
	my $output_evalue=File::Spec->catfile($output, "evalue.csv");
	open( EVALUE, ">$output_evalue" ) or die "Cannot open $output_evalue\n$!\n";
	for my $f ( keys %hits ) {
		print EVALUE ( $f . "," );
		foreach my $e ( @{ $hits{$f}->{EVALUE} } ) {
			if ( defined $e ) {
				if ( $e + 0 == 0 ) {
					print EVALUE "99999";
				}
				else {
					print EVALUE abs( log($e) / log(10) ) . ",";
				}
			}
			print EVALUE ",";

		}
		print EVALUE "\n";
	}
	close(EVALUE);
}

sub printRank {
	my $output_rank=File::Spec->catfile($output, "rank.csv");
	open( RANKING, ">$output_rank" ) or die "Cannot open $output_rank\n$!\n";
	for my $f ( keys %hits ) {
		print RANKING ( $f . "," );
		foreach my $r ( @{ $hits{$f}->{RANK} } ) {
			print RANKING $r if ( defined $r );
			print RANKING ",";
		}
		print RANKING "\n";
	}
	close(RANKING);
}

sub printFrame {
	my $output_frame=File::Spec->catfile($output, "frame.csv");
	open( FRAMES, ">$output_frame" ) or die "Cannot open $output_frame\n$!\n";
	for my $f ( keys %hits ) {
		print FRAMES ( $f . "," );
		foreach my $r ( @{ $hits{$f}->{FRAME} } ) {
			print FRAMES "[$r]" if(defined $r);
			print FRAMES ",";
		}
		print FRAMES "\n";
	}
	close(FRAMES);
}

#################################################################
#                            STATISTICS
#################################################################

# scalar @list is always > 0
sub median {
	my @list = @_;
	my $len  = scalar @list;
	return $list[ int($len / 2) ];
}

sub std {
	my ( $mean, @values ) = @_;
	my $var = variance( $mean, @values );
	return "NaN" if ( $var eq "NaN" );
	return sqrt($var);
}

# variance = sum((Xi - mean)^2) / (n - 1)
sub variance {
	my ( $mean, @values ) = @_;
	my $length = scalar @values;
	my $var    = "NaN";
	if ( $length == 1 ) {
		$var = 0;
	}
	elsif ( $length > 1 ) {
		my $accum = 0;
		map { $accum += ( $_ - $mean )**2; } @values;
		$var = $accum / ( $length - 1 );    
	}
	return $var;
}
# West's algorithm for computing variance
# NOT USED
sub var{
	my ( $mean, @values ) = @_;
	my $length = scalar @values;
	my $var    = "NaN";
	if ( $length == 1 ) {
		$var = 0;
	}
	elsif ( $length > 1 ) {
		my $accum = 0.0;
    	my $accum2 = 0.0;
    	for (my $i = 0; $i < $length; $i++) {
    		$accum += ($values[$i] - $mean)**2;
     		$accum2 += ($values[$i] - $mean);
    	}
    	$var = ($accum - (($accum2**2) / $length)) / ($length - 1);
	}
    return $var;
}




sub mean {
	my @values = @_;
	my $sum    = 0;
	map { $sum += $_ } @values;
	my $mean = $sum / ( $#values + 1 );
	return $mean;
}

sub interactive{	
	$input = prompt("Enter the name of the file containing the BLAST result");
	die "$input does not exist" unless(-e $input);
	
	$output = prompt("Enter the name of the directory where the results will be stored","$input" . "_STAT");
	print "$output\n";
	
	$threshold = prompt("Define a threshold (OPTIONAL)" );
	print "$program\n";
}

sub prompt($$) {
	my ( $text, $default ) = @_;
    if ($default) {print $text, "[", $default, "]:\n";}
    else {print $text, ":\n";}
    
	$| = 1;
	$_ = <STDIN>;

	s/\r?\n//g;

	if ( defined $default ) { return $_ ? $_ : $default;}
	else {return $_;}
}

#################################################################
#                            HELP
#################################################################
sub help {
	my ($message) = @_;
	print "\n$message\n\n" if ( defined $message );
	print "Name
    $0 - Parse a blast file generated from SWeBLAST.pl and give some statistics (ranking, e-value...) about the file. 
Synopsis
    $0 [OPTION]...
Description
   -help or -h :      Show this message.
   -input or -i :     Input file.
   -output or -o :    Output directory to be created with the results files (OPTIONAL).
   -threshold or -t:  Hits with an e-value above the specified value will be discarded (OPTIONAL).

Output
   summary.csv :      Display the results
   stat.csv :         Display some stats on the log of the e-value and rank (min,max,median,mean,standard deviation)
   evalue.csv :       Display the log of the e-value
   rank.csv :         Display the rank of the sequences
   frame.csv :        Display the the orientation of the ORF (+1,+2,+3,-1,-2,-3)

Exemple
   perl $0 -i blast1.txt -o /home/user/project1 [-t 0.0001]\n";
	exit 0;
}
